import time
import logging
from collections import defaultdict
from requests.models import CaseInsensitiveDict

logger = logging.getLogger('pydispix')


class RateLimitedEndpoint:
    def __init__(self, default_delay: int = 0):
        self.rate_limited = True            # Not all endpoints are rate-limited
        self.requests_limit = None          # Total number of requests before reset time wait
        self.remaining_requests = 1          # Remaining number of requests before reset time wait
        self.reset_time = 0                 # How much time to wait once we hit reset
        self.cooldown_time = 0              # Some endpoints force longer cooldown times
        self.default_delay = default_delay  # If no other limit is found, how long should we wait

    def update_from_headers(self, headers: CaseInsensitiveDict[str]):
        if 'requests-remaining' not in headers:
            self.rate_limited = False
            return

        self.remaining_requests = int(headers.get('requests-remaining', 1))
        self.reset_time = int(headers.get('requests-reset', 0))
        self.cooldown_time = int(headers.get('cooldown-reset', 0))
        if "requests-limit" in headers:
            self.requests_limit = int(headers["requests-limit"])

    def get_wait_time(self):
        if not self.rate_limited:
            return self.default_delay
        if self.cooldown_time != 0:
            return self.cooldown_time
        if self.remaining_requests == 0:
            return self.reset_time

        return self.default_delay

    def wait(self):
        if not self.rate_limited:
            logger.debug("Sleeping default delay, not rate limited.")
            return time.sleep(self.default_delay)
        if self.cooldown_time != 0:
            logger.warning(f"Sleeping {self.cooldown_time}s, on cooldown.")
            return time.sleep(self.cooldown_time)
        if self.remaining_requests == 0:
            logger.warning(f"Sleeping {self.reset_time}s, on reset.")
            return time.sleep(self.reset_time)

        logger.debug(f"Sleeping default delay, {self.remaining_requests} requests remaining.")
        return time.sleep(self.default_delay)


class RateLimiter:
    def __init__(self):
        self.rate_limits = defaultdict(RateLimitedEndpoint)

    def update_from_headers(self, endpoint: str, headers: CaseInsensitiveDict[str]):
        limiter = self.rate_limits[endpoint]
        limiter.update_from_headers(headers)

    def wait(self, endpoint: str):
        limiter = self.rate_limits[endpoint]
        limiter.wait()



from pydispix.client import Client  # noqa
from pydispix.canvas import Canvas, Pixel  # noqa
from pydispix.color import Color, Colour, parse_color, parse_colour  # noqa
from pydispix.ratelimits import RateLimiter, RateLimitedEndpoint  # noqa
from pydispix.log import setup_logging

setup_logging()



import sys
import os
import logging


def setup_logging():
    DEBUG_MODE = 'DEBUG' in os.environ

    log_format = logging.Formatter("%(asctime)s | %(name)s | %(levelname)s | %(message)s")
    stream_handler = logging.StreamHandler(stream=sys.stdout)
    stream_handler.setFormatter(log_format)

    logger = logging.getLogger("pydispix")
    logger.setLevel(logging.DEBUG if DEBUG_MODE else logging.INFO)
    logger.addHandler(stream_handler)



"""Tool for automatically drawing images."""
import logging

import PIL.Image

from pydispix.canvas import Pixel
from pydispix.client import Client


logger = logging.getLogger('pydispix')


class AutoDrawer:
    """Tool for automatically drawing images."""

    def __init__(
        self,
        client: Client,
        x: int, y: int,
        grid: list[list[Pixel]]
    ):
        """Store the plan."""
        self.client = client
        self.grid = grid
        # Top left coords.
        self.x0 = x
        self.y0 = y
        # Bottom right coords.
        self.x1 = x + len(self.grid[0])
        self.y1 = y + len(self.grid)

    @classmethod
    def load_image(
        cls,
        client: Client,
        xy: tuple[int, int],
        image: PIL.Image.Image,
        scale: float = 1
    ) -> 'AutoDrawer':
        """Draw from the pixels of an image."""
        if image.mode == 'RGBA':
            new_image = PIL.Image.new('RGB', image.size)
            new_image.paste(image, mask=image)
            image = new_image

        width = round(image.width * scale)
        height = round(image.height * scale)
        data = list(image.resize((width, height)).getdata())
        grid = [
            [Pixel(*pixel) for pixel in data[start:start + width]]
            for start in range(0, len(data), width)
        ]
        return cls(client, *xy, grid)

    @classmethod
    def load(cls, client: Client, data: str) -> 'AutoDrawer':
        """Draw from a string that specifies the pixels.
        `data` should be a multi-line string. The first two lines are the x
        and y coordinates of the top left of the image to draw. The second two
        are the width and height of the image. The rest of the lines are the
        pixels of the image, as hex codes (horizontal scanlines, left-to-right
        top-to-bottom).
        """
        lines = data.split('\n')
        x = int(lines.pop(0))
        y = int(lines.pop(0))
        width = int(lines.pop(0))
        height = int(lines.pop(0))
        grid = []
        for _ in range(height):
            row = []
            for _ in range(width):
                row.append(Pixel.from_hex(lines.pop(0)))
            grid.append(row)
        return cls(client, x, y, grid)

    def draw(self):
        """Draw the pixels of the image."""
        canvas = self.client.get_canvas()
        for x in range(self.x0, self.x1):
            for y in range(self.y0, self.y1):
                dx = x - self.x0
                dy = y - self.y0
                colour = self.grid[dy][dx]
                if canvas[x, y] == colour:
                    logger.debug(
                        f'Skipping already correct pixel at {x}, {y}.'
                    )
                    continue
                self.client.put_pixel(x, y, colour)
                canvas = self.client.get_canvas()



import os
import requests
import logging
from collections import namedtuple
from typing import Union, Optional

from pydispix.ratelimits import RateLimiter
from pydispix.canvas import Canvas, Pixel
from pydispix.color import Color, parse_color

logger = logging.getLogger('pydispix')
Dimensions = namedtuple('Dimensions', ('width', 'height'))


class Client:
    """HTTP client to the pixel API."""
    def __init__(self, token: Optional[str] = None, base_url="https://pixels.pythondiscord.com"):
        if token is None:
            try:
                token = os.environ['TOKEN']
            except KeyError:
                raise RuntimeError("Unable to load token, 'TOKEN' environmental variable not found.")

        self.token = token
        self.base_url = base_url
        self.headers = {
            "Authorization": "Bearer " + token,
            "User-Agent": "ItsDrike pydispix",
        }
        self.rate_limiter = RateLimiter()
        self.width, self.height = self.size = self.get_dimensions()

    def make_request(
        self,
        method: str,
        endpoint_url: str,
        *,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
        parse_json: bool = True,
        ratelimit_after: bool = False
    ) -> Union[bytes, dict]:
        if not endpoint_url.startswith("/"):
            endpoint_url = "/" + endpoint_url

        if not ratelimit_after:
            self.rate_limiter.wait(endpoint_url)

        logger.debug(f'Request: {method} {endpoint_url} data={data!r} params={params!r}.')

        response = requests.request(method, self.base_url + endpoint_url, json=data, headers=self.headers, params=params)
        self.rate_limiter.update_from_headers(endpoint_url, response.headers)
        if ratelimit_after:
            self.rate_limiter.wait(endpoint_url)

        if parse_json:
            return response.json()
        else:
            return response.content

    def get_dimensions(self) -> Dimensions:
        data = self.make_request("GET", "get_size")
        return Dimensions(width=data["width"], height=data["height"])

    def get_canvas(self) -> Canvas:
        data = self.make_request("GET", "get_pixels", parse_json=False)
        return Canvas(self.size, data)

    def get_pixel(self, x: int, y: int) -> Pixel:
        data = self.make_request("GET", "get_pixel", params={"x": x, "y": y})
        hex_color = data["rgb"]
        return Pixel.from_hex(hex_color)

    def put_pixel(self, x: int, y: int, color: Union[int, str, Color]) -> str:
        """Draw a pixel and return a message."""
        # Wait for ratelimits *after* making request, not before. This makes
        # sense because we don't know how the canvas may have changed by the
        # time we have finished waiting, whereas for GET endpoints, we want to
        # return the information as soon as it is given.
        data = self.make_request(
            'POST', 'set_pixel',
            data={
                'x': x,
                'y': y,
                'rgb': parse_color(color)
            },
            ratelimit_after=True
        )
        logger.info('Success: {message}'.format(**data))
        return data['message']
